{$INCLUDE cHeader.inc}
unit cUtil;

interface

uses
  SysUtils,
  Math;



{                                                                              }
{ Miscellaneous utility functions v0.05 (L0)                                   }
{                                                                              }
{   A collection of string functions.                                          }
{                                                                              }
{                                                                              }
{ This unit is copyright  2000 by David Butler (david@e.co.za)                }
{                                                                              }
{ This unit is part of Delphi Fundamentals, it's original name is cUtil.       }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   2000/02/02  v0.01  Initial version                                         }
{   2000/03/08  v0.02  Moved RealArray / IntegerArray functions from cMaths.   }
{   2000/04/10  v0.03  Added Append, Renamed Delete to Remove and added        }
{                      StringArrays.                                           }
{   2000/05/03  v0.04  Added Path functions.                                   }
{                      Added locked integer manipulation functions (by         }
{                      Azret Botash).                                          }
{   2000/05/08  v0.05  Cleaned up unit.                                        }
{                      188 lines interface. 1171 lines implementation.         }
{                                                                              }



{ Swap                                                                         }
Procedure Swap (var X, Y : Boolean); overload;
Procedure Swap (var X, Y : Byte); overload;
Procedure Swap (var X, Y : Word); overload;
Procedure Swap (var X, Y : Integer); overload;
Procedure Swap (var X, Y : Int64); overload;
Procedure Swap (var X, Y : Single); overload;
Procedure Swap (var X, Y : Extended); overload;
Procedure Swap (var X, Y : Double); overload;
Procedure Swap (var X, Y : String); overload;
Procedure Swap (var X, Y : Pointer); overload;
Procedure Swap (var X, Y : TObject); overload;

{ Cond                                                                         }
Function  Cond (const Expr : Boolean; const TrueValue, FalseValue : Int64) : Int64; overload;
Function  Cond (const Expr : Boolean; const TrueValue, FalseValue : Extended) : Extended; overload;
Function  Cond (const Expr : Boolean; const TrueValue, FalseValue : String) : String; overload;
Function  Cond (const Expr : Boolean; const TrueValue, FalseValue : TObject) : TObject; overload;

{ Floating point                                                               }
type
  TFPUPrecision = (pSingle, pDouble, pExtended);

Procedure SetFPUPrecision (const Precision : TFPUPrecision);
Function  FloatEqual (const X, Y : Extended) : Boolean;
Function  FloatZero (const X : Extended) : Boolean;
Procedure FloatDefuzz (var X : Single); overload;
Procedure FloatDefuzz (var X : Double); overload;
Procedure FloatDefuzz (var X : Extended); overload;

var
  FloatCompareTolerance : Extended = 1e-10;

{ Character sets                                                               }
type
  CharSet = Set of Char;

{ Dynamic Arrays                                                               }
type
  ByteArray     = Array of Byte;
  IntegerArray  = Array of Integer;
  Int64Array    = Array of Int64;
  SingleArray   = Array of Single;
  DoubleArray   = Array of Double;
  ExtendedArray = Array of Extended;
  StringArray   = Array of String;
  ObjectArray   = Array of TObject;
  CharSetArray  = Array of CharSet;

Function  Append (var V : ByteArray; const R : Byte) : Integer; overload;
Function  Append (var V : IntegerArray; const R : Integer) : Integer; overload;
Function  Append (var V : Int64Array; const R : Int64) : Integer; overload;
Function  Append (var V : SingleArray; const R : Single) : Integer; overload;
Function  Append (var V : DoubleArray; const R : Double) : Integer; overload;
Function  Append (var V : ExtendedArray; const R : Extended) : Integer; overload;
Function  Append (var V : StringArray; const R : String) : Integer; overload;
Function  Append (var V : ObjectArray; const R : TObject) : Integer; overload;

Function  Append (var S : ByteArray; const V : ByteArray) : Integer; overload;
Function  Append (var S : IntegerArray; const V : IntegerArray) : Integer; overload;
Function  Append (var S : Int64Array; const V : Int64Array) : Integer; overload;
Function  Append (var S : SingleArray; const V : SingleArray) : Integer; overload;
Function  Append (var S : DoubleArray; const V : DoubleArray) : Integer; overload;
Function  Append (var S : ExtendedArray; const V : ExtendedArray) : Integer; overload;
Function  Append (var S : StringArray; const V : StringArray) : Integer; overload;
Function  Append (var S : ObjectArray; const V : ObjectArray) : Integer; overload;

Procedure Remove (var V : ByteArray; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : IntegerArray; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : Int64Array; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : SingleArray; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : DoubleArray; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : ExtendedArray; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : StringArray; const Idx : Integer; const Count : Integer = 1); overload;
Procedure Remove (var V : ObjectArray; const Idx : Integer; const Count : Integer = 1;
          const FreeObjects : Boolean = False); overload;

Procedure FreeObjectArray (var V);

Function  PosNext (const Find : Integer; const V : IntegerArray; const PrevPos : Integer = -1;
          const IsSortedAscending : Boolean = False) : Integer; overload;
Function  PosNext (const Find : Int64; const V : Int64Array; const PrevPos : Integer = -1;
          const IsSortedAscending : Boolean = False) : Integer; overload;
Function  PosNext (const Find : Single; const V : SingleArray; const PrevPos : Integer = -1;
          const IsSortedAscending : Boolean = False) : Integer; overload;
Function  PosNext (const Find : Double; const V : DoubleArray; const PrevPos : Integer = -1;
          const IsSortedAscending : Boolean = False) : Integer; overload;
Function  PosNext (const Find : Extended; const V : ExtendedArray; const PrevPos : Integer = -1;
          const IsSortedAscending : Boolean = False) : Integer; overload;
Function  PosNext (const Find : String; const V : StringArray; const PrevPos : Integer = -1;
          const IsSortedAscending : Boolean = False) : Integer; overload;

Function  Intersection (const V1, V2 : IntegerArray; const IsSortedAscending : Boolean = False) : IntegerArray; overload;
Function  Intersection (const V1, V2 : Int64Array; const IsSortedAscending : Boolean = False) : Int64Array; overload;
Function  Intersection (const V1, V2 : SingleArray; const IsSortedAscending : Boolean = False) : SingleArray; overload;
Function  Intersection (const V1, V2 : DoubleArray; const IsSortedAscending : Boolean = False) : DoubleArray; overload;
Function  Intersection (const V1, V2 : ExtendedArray; const IsSortedAscending : Boolean = False) : ExtendedArray; overload;
Function  Intersection (const V1, V2 : StringArray; const IsSortedAscending : Boolean = False) : StringArray; overload;

Procedure Reverse (var V : ByteArray); overload;
Procedure Reverse (var V : IntegerArray); overload;
Procedure Reverse (var V : Int64Array); overload;
Procedure Reverse (var V : SingleArray); overload;
Procedure Reverse (var V : DoubleArray); overload;
Procedure Reverse (var V : ExtendedArray); overload;
Procedure Reverse (var V : StringArray); overload;
Procedure Reverse (var V : ObjectArray); overload;

Function  AsIntegerArray (const V : Array of Integer) : IntegerArray;
Function  AsInt64Array (const V : Array of Int64) : Int64Array;
Function  AsSingleArray (const V : Array of Extended) : SingleArray;
Function  AsDoubleArray (const V : Array of Extended) : DoubleArray;
Function  AsExtendedArray (const V : Array of Extended) : ExtendedArray;
Function  AsStringArray (const V : Array of String) : StringArray;
Function  AsCharSetArray (const V : Array of CharSet) : CharSetArray;

{ Type conversion                                                              }
Function  StrToIntDefault (const S : String; const Default : Integer = 0) : Integer;
Function  StrToFloatDefault (const S : String; const Default : Extended = 0) : Extended;
Function  BooleanToStr (const B : Boolean) : String;
Function  StrToBoolean (const S : String) : Boolean;
Function  ExceptionToStr (const E : Exception) : String;

{ Max values                                                                   }
const
  MaxByte     = High (Byte);
  MaxWord     = High (Word);
  MaxLongWord = High (LongWord);
  MaxInteger  = High (Integer);
  MaxInt64    = High (Int64);

{ Real / Integer types for mathematical operations                             }
type
  TReal         = Extended;
  TInteger      = Int64;
  TRealArray    = ExtendedArray;
  TIntegerArray = Int64Array;

const
  MaxTReal    = MaxExtended;
  MaxTInteger = High (TInteger);

{ Path functions                                                               }
const
  PathSeparator = {$IFDEF UNIX}'/'{$ELSE}'\'{$ENDIF};

Function UnixPathToWindowsPath (const Path : String) : String;
Function WindowsPathToUnixPath (const Path : String) : String;
Function PathWithSlashAtEnd (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathWithoutSlashAtEnd (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathLeftElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathRightElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathWithoutLeftElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathWithoutRightElement (const Path : String; const SlashCh : Char = PathSeparator) : String;

{ Locked integer manipulation (simple synchronization)                         }
Function LockedAdd (var Target : Integer; const Value : Integer) : Integer; assembler; register;
Function LockedCompareExchange (var Target : Integer; const Exch, Comp : Integer) : Integer; assembler; register;
Function LockedDec (var Target : Integer) : Integer; assembler; register;
Function LockedExchange (var Target : Integer; Value : Integer) : Integer; assembler; register;
Function LockedExchangeAdd (var Target : Integer; Value : Integer) : Integer; assembler; register;
Function LockedExchangeDec (var Target : Integer) : Integer; assembler; register;
Function LockedExchangeInc (var Target : Integer) : Integer; assembler; register;
Function LockedExchangeSub (var Target : Integer; Value : Integer) : Integer; assembler; register;
Function LockedInc (var Target : Integer) : Integer; assembler; register;
Function LockedSub (var Target : Integer; const Value : Integer) : Integer; assembler; register;



implementation

uses
  // Delphi Fundamentals (L0)
  cStrings;      // Copy functions



{                                                                              }
{ Swap                                                                         }
{                                                                              }
Procedure Swap (var X, Y : Boolean);
var F : Boolean;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Byte);
var F : Byte;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Word);
var F : Word;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Integer);
var F : Integer;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Int64);
var F : Int64;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Extended);
var F : Extended;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Double);
var F : Double;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Single);
var F : Single;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : String);
var F : String;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : Pointer);
var F : Pointer;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;

Procedure Swap (var X, Y : TObject);
var F : TObject;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;



{                                                                              }
{ Cond                                                                         }
{                                                                              }
Function Cond (const Expr : Boolean; const TrueValue, FalseValue : Int64) : Int64;
  Begin
    if Expr then
      Result := TrueValue else
      Result := FalseValue;
  End;

Function Cond (const Expr : Boolean; const TrueValue, FalseValue : Extended) : Extended;
  Begin
    if Expr then
      Result := TrueValue else
      Result := FalseValue;
  End;

Function Cond (const Expr : Boolean; const TrueValue, FalseValue : String) : String;
  Begin
    if Expr then
      Result := TrueValue else
      Result := FalseValue;
  End;

Function Cond (const Expr : Boolean; const TrueValue, FalseValue : TObject) : TObject; overload;
  Begin
    if Expr then
      Result := TrueValue else
      Result := FalseValue;
  End;



{                                                                              }
{ Floating point                                                               }
{                                                                              }

{ You can modify the level of accuracy by changing the FPU's control word.     }
{ The default accuracy, as initialized by the Delphi runtime library,          }
{ is the slowest, but most precise one (i.e. extended). Note that changing     }
{ this only changes the execution time of division and, in the case of         }
{ Pentium II and Pentium III processors, square roots.                         }
{ See http://econos.com/optimize/float.htm#ctrlwrd                             }
Procedure SetFPUPrecision (const Precision : TFPUPrecision);
  Begin
    Case Precision of
      pSingle   : Set8087CW (Default8087CW and $FCFF);
      pDouble   : Set8087CW ((Default8087CW and $FCFF) or $0200);
      pExtended : Set8087CW (Default8087CW or $0300);
    end;
  End;

Function FloatEqual (const X, Y : Extended) : Boolean;
  Begin
    Result := Abs (X - Y) <= FloatCompareTolerance;
  End;

Function FloatZero (const X : Extended) : Boolean;
  Begin
    Result := Abs (X) <= FloatCompareTolerance;
  End;

{ FloatDefuzz returns zero if the number is nearly zero. Avoids propogation    }
{   of nearly-zero values. The term 'fuzz' is courtesy of Earl F. Glynn        }
{   (EarlGlynn@att.net) which adapted it from the APL language.                }
Procedure FloatDefuzz (var X : Single);
  Begin
    if Abs (X) <= FloatCompareTolerance then
      X := 0.0;
  End;

Procedure FloatDefuzz (var X : Double);
  Begin
    if Abs (X) <= FloatCompareTolerance then
      X := 0.0;
  End;

Procedure FloatDefuzz (var X : Extended);
  Begin
    if Abs (X) <= FloatCompareTolerance then
      X := 0.0;
  End;



{                                                                              }
{ Append                                                                       }
{                                                                              }
Function Append (var V : ByteArray; const R : Byte) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : IntegerArray; const R : Integer) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : Int64Array; const R : Int64) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : SingleArray; const R : Single) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : DoubleArray; const R : Double) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : ExtendedArray; const R : Extended) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : StringArray; const R : String) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var V : ObjectArray; const R : TObject) : Integer;
  Begin
    Result := Length (V);
    SetLength (V, Result + 1);
    V [Result] := R;
  End;

Function Append (var S : StringArray; const V : StringArray) : Integer;
var I, LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        For I := 0 to LV - 1 do
          S [Result + I] := V [I];
      end;
  End;

Function Append (var S : ByteArray; const V : ByteArray) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;

Function Append (var S : IntegerArray; const V : IntegerArray) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;

Function Append (var S : Int64Array; const V : Int64Array) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;

Function Append (var S : SingleArray; const V : SingleArray) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;

Function Append (var S : DoubleArray; const V : DoubleArray) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;

Function Append (var S : ExtendedArray; const V : ExtendedArray) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;

Function Append (var S : ObjectArray; const V : ObjectArray) : Integer;
var LV : Integer;
  Begin
    Result := Length (S);
    LV := Length (V);
    if LV > 0 then
      begin
        SetLength (S, Result + LV);
        Move (V [0], S [Result], Sizeof (V [0]) * LV);
      end;
  End;



{                                                                              }
{ Remove                                                                       }
{                                                                              }
Procedure Remove (var V : ByteArray; const Idx : Integer; const Count : Integer);
var I, J, L, M : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    M := L - J - I;
    if M > 0 then
      Move (V [I + J], V [I], M * SizeOf (V [0]));
    SetLength (V, L - J);
  End;

Procedure Remove (var V : IntegerArray; const Idx : Integer; const Count : Integer);
var I, J, L, M : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    M := L - J - I;
    if M > 0 then
      Move (V [I + J], V [I], M * SizeOf (V [0]));
    SetLength (V, L - J);
  End;

Procedure Remove (var V : Int64Array; const Idx : Integer; const Count : Integer);
var I, J, L, M : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    M := L - J - I;
    if M > 0 then
      Move (V [I + J], V [I], M * SizeOf (V [0]));
    SetLength (V, L - J);
  End;

Procedure Remove (var V : SingleArray; const Idx : Integer; const Count : Integer);
var I, J, L, M : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    M := L - J - I;
    if M > 0 then
      Move (V [I + J], V [I], M * SizeOf (V [0]));
    SetLength (V, L - J);
  End;

Procedure Remove (var V : DoubleArray; const Idx : Integer; const Count : Integer);
var I, J, L, M : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    M := L - J - I;
    if M > 0 then
      Move (V [I + J], V [I], M * SizeOf (V [0]));
    SetLength (V, L - J);
  End;

Procedure Remove (var V : ExtendedArray; const Idx : Integer; const Count : Integer);
var I, J, L, M : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    M := L - J - I;
    if M > 0 then
      Move (V [I + J], V [I], M * SizeOf (V [0]));
    SetLength (V, L - J);
  End;

Procedure Remove (var V : ObjectArray; const Idx : Integer; const Count : Integer; const FreeObjects : Boolean);
var I, J, K, L : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    if FreeObjects then
      For K := I to L - J - 1 do
        FreeAndNil (V [K]);
    SetLength (V, L - J);
  End;

Procedure Remove (var V : StringArray; const Idx : Integer; const Count : Integer);
var I, J, K, L : Integer;
  Begin
    L := Length (V);
    if (Idx >= L) or (Idx + Count <= 0) or (L = 0) or (Count = 0) then
      exit;
    I := Max (Idx, 0);
    J := Min (Count, L - I);
    For K := I to L - J - 1 do
      V [K] := V [K + J];
    SetLength (V, L - J);
  End;


Procedure FreeObjectArray (var V);
var I : Integer;
    A : ObjectArray absolute V;
  Begin
    For I := Length (A) - 1 downto 0 do
      FreeAndNil (A [I]);
  End;



{                                                                              }
{ PosNext                                                                      }
{   PosNext finds the next occurance of Find in V, -1 if it was not found.     }
{     Searches from item [PrevPos + 1], ie PrevPos = -1 to find first          }
{     occurance.                                                               }
{                                                                              }
Function PosNext (const Find : Single; const V : SingleArray; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Single;
  Begin
    if IsSortedAscending then
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first, binary search
          begin
            L := 0;
            H := Length (V) - 1;
            Repeat
              I := (L + H) div 2;
              D := V [I];
              if FloatEqual (D, Find) then
                begin
                  While (I > 0) and FloatEqual (V [I - 1], Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Length (V) - 1 then
            Result := -1 else
            if FloatEqual (V [PrevPos + 1], Find) then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Length (V) - 1 do
          if FloatEqual (V [I], Find) then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;

Function PosNext (const Find : Double; const V : DoubleArray; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Double;
  Begin
    if IsSortedAscending then
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first, binary search
          begin
            L := 0;
            H := Length (V) - 1;
            Repeat
              I := (L + H) div 2;
              D := V [I];
              if FloatEqual (D, Find) then
                begin
                  While (I > 0) and FloatEqual (V [I - 1], Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Length (V) - 1 then
            Result := -1 else
            if FloatEqual (V [PrevPos + 1], Find) then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Length (V) - 1 do
          if FloatEqual (V [I], Find) then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;

Function PosNext (const Find : Extended; const V : ExtendedArray; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Extended;
  Begin
    if IsSortedAscending then
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first, binary search
          begin
            L := 0;
            H := Length (V) - 1;
            Repeat
              I := (L + H) div 2;
              D := V [I];
              if FloatEqual (D, Find) then
                begin
                  While (I > 0) and FloatEqual (V [I - 1], Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Length (V) - 1 then
            Result := -1 else
            if FloatEqual (V [PrevPos + 1], Find) then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Length (V) - 1 do
          if FloatEqual (V [I], Find) then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;

Function PosNext (const Find : Integer; const V : IntegerArray; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Integer;
  Begin
    if IsSortedAscending then
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first, binary search
          begin
            L := 0;
            H := Length (V) - 1;
            Repeat
              I := (L + H) div 2;
              D := V [I];
              if D = Find then
                begin
                  While (I > 0) and (V [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Length (V) - 1 then
            Result := -1 else
            if V [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else // linear search
      begin
        For I := Max (PrevPos + 1, 0) to Length (V) - 1 do
          if V [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;

Function PosNext (const Find : Int64; const V : Int64Array; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Int64;
  Begin
    if IsSortedAscending then
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first, binary search
          begin
            L := 0;
            H := Length (V) - 1;
            Repeat
              I := (L + H) div 2;
              D := V [I];
              if D = Find then
                begin
                  While (I > 0) and (V [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Length (V) - 1 then
            Result := -1 else
            if V [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else // linear search
      begin
        For I := Max (PrevPos + 1, 0) to Length (V) - 1 do
          if V [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;

Function PosNext (const Find : String; const V : StringArray; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : String;
  Begin
    if IsSortedAscending then
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first, binary search
          begin
            L := 0;
            H := Length (V) - 1;
            Repeat
              I := (L + H) div 2;
              D := V [I];
              if D = Find then
                begin
                  While (I > 0) and (V [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Length (V) - 1 then
            Result := -1 else
            if V [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else // linear search
      begin
        For I := Max (PrevPos + 1, 0) to Length (V) - 1 do
          if V [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;



{                                                                              }
{ Intersection                                                                 }
{                                                                              }
Function Intersection (const V1, V2 : SingleArray; const IsSortedAscending : Boolean) : SingleArray;
var I, J, L, LV : Integer;
  Begin
    SetLength (Result, 0);
    if IsSortedAscending then
      begin
        I := 0;
        J := 0;
        L := Length (V1);
        LV := Length (V2);
        While (I < L) and (J < LV) do
          begin
            While (I < L) and (V1 [I] < V2 [J]) do
              Inc (I);
            if I < L then
              begin
                if FloatEqual (V1 [I], V2 [J]) then
                  Append (Result, V1 [I]);
                While (J < LV) and (FloatEqual (V2 [J], V1 [I]) or (V2 [J] < V1 [I])) do
                  Inc (J);
              end;
          end;
      end else
      For I := 0 to Length (V1) - 1 do
        if (PosNext (V1 [I], V2) >= 0) and (PosNext (V1 [I], Result) = -1) then
          Append (Result, V1 [I]);
  End;

Function Intersection (const V1, V2 : DoubleArray; const IsSortedAscending : Boolean) : DoubleArray;
var I, J, L, LV : Integer;
  Begin
    SetLength (Result, 0);
    if IsSortedAscending then
      begin
        I := 0;
        J := 0;
        L := Length (V1);
        LV := Length (V2);
        While (I < L) and (J < LV) do
          begin
            While (I < L) and (V1 [I] < V2 [J]) do
              Inc (I);
            if I < L then
              begin
                if FloatEqual (V1 [I], V2 [J]) then
                  Append (Result, V1 [I]);
                While (J < LV) and (FloatEqual (V2 [J], V1 [I]) or (V2 [J] < V1 [I])) do
                  Inc (J);
              end;
          end;
      end else
      For I := 0 to Length (V1) - 1 do
        if (PosNext (V1 [I], V2) >= 0) and (PosNext (V1 [I], Result) = -1) then
          Append (Result, V1 [I]);
  End;

Function Intersection (const V1, V2 : ExtendedArray; const IsSortedAscending : Boolean) : ExtendedArray;
var I, J, L, LV : Integer;
  Begin
    SetLength (Result, 0);
    if IsSortedAscending then
      begin
        I := 0;
        J := 0;
        L := Length (V1);
        LV := Length (V2);
        While (I < L) and (J < LV) do
          begin
            While (I < L) and (V1 [I] < V2 [J]) do
              Inc (I);
            if I < L then
              begin
                if FloatEqual (V1 [I], V2 [J]) then
                  Append (Result, V1 [I]);
                While (J < LV) and (FloatEqual (V2 [J], V1 [I]) or (V2 [J] < V1 [I])) do
                  Inc (J);
              end;
          end;
      end else
      For I := 0 to Length (V1) - 1 do
        if (PosNext (V1 [I], V2) >= 0) and (PosNext (V1 [I], Result) = -1) then
          Append (Result, V1 [I]);
  End;

Function Intersection (const V1, V2 : IntegerArray; const IsSortedAscending : Boolean) : IntegerArray;
var I, J, L, LV : Integer;
  Begin
    SetLength (Result, 0);
    if IsSortedAscending then
      begin
        I := 0;
        J := 0;
        L := Length (V1);
        LV := Length (V2);
        While (I < L) and (J < LV) do
          begin
            While (I < L) and (V1 [I] < V2 [J]) do
              Inc (I);
            if I < L then
              begin
                if V1 [I] = V2 [J] then
                  Append (Result, V1 [I]);
                While (J < LV) and (V2 [J] <= V1 [I]) do
                  Inc (J);
              end;
          end;
      end else
      For I := 0 to Length (V1) - 1 do
        if (PosNext (V1 [I], V2) >= 0) and (PosNext (V1 [I], Result) = -1) then
          Append (Result, V1 [I]);
  End;

Function Intersection (const V1, V2 : Int64Array; const IsSortedAscending : Boolean) : Int64Array;
var I, J, L, LV : Integer;
  Begin
    SetLength (Result, 0);
    if IsSortedAscending then
      begin
        I := 0;
        J := 0;
        L := Length (V1);
        LV := Length (V2);
        While (I < L) and (J < LV) do
          begin
            While (I < L) and (V1 [I] < V2 [J]) do
              Inc (I);
            if I < L then
              begin
                if V1 [I] = V2 [J] then
                  Append (Result, V1 [I]);
                While (J < LV) and (V2 [J] <= V1 [I]) do
                  Inc (J);
              end;
          end;
      end else
      For I := 0 to Length (V1) - 1 do
        if (PosNext (V1 [I], V2) >= 0) and (PosNext (V1 [I], Result) = -1) then
          Append (Result, V1 [I]);
  End;

Function Intersection (const V1, V2 : StringArray; const IsSortedAscending : Boolean) : StringArray;
var I, J, L, LV : Integer;
  Begin
    SetLength (Result, 0);
    if IsSortedAscending then
      begin
        I := 0;
        J := 0;
        L := Length (V1);
        LV := Length (V2);
        While (I < L) and (J < LV) do
          begin
            While (I < L) and (V1 [I] < V2 [J]) do
              Inc (I);
            if I < L then
              begin
                if V1 [I] = V2 [J] then
                  Append (Result, V1 [I]);
                While (J < LV) and (V2 [J] <= V1 [I]) do
                  Inc (J);
              end;
          end;
      end else
      For I := 0 to Length (V1) - 1 do
        if (PosNext (V1 [I], V2) >= 0) and (PosNext (V1 [I], Result) = -1) then
          Append (Result, V1 [I]);
  End;



{                                                                              }
{ Reverse                                                                      }
{                                                                              }
Procedure Reverse (var V : ByteArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : IntegerArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : Int64Array);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : SingleArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : DoubleArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : ExtendedArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : StringArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

Procedure Reverse (var V : ObjectArray);
var I : Integer;
  Begin
    For I := 1 to Length (V) div 2 do
      Swap (V [I - 1], V [Length (V) - I]);
  End;

{                                                                              }
{ Returns an open array (V) as a dynamic array.                                }
{                                                                              }
Function  AsIntegerArray (const V : Array of Integer) : IntegerArray;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;

Function  AsInt64Array (const V : Array of Int64) : Int64Array;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;

Function  AsSingleArray (const V : Array of Extended) : SingleArray;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;

Function  AsDoubleArray (const V : Array of Extended) : DoubleArray;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;

Function  AsExtendedArray (const V : Array of Extended) : ExtendedArray;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;

Function  AsStringArray (const V : Array of String) : StringArray;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;

Function  AsCharSetArray (const V : Array of CharSet) : CharSetArray;
var I : Integer;
  Begin
    SetLength (Result, High (V) + 1);
    For I := 0 to High (V) do
     Result [I] := V [I];
  End;



{                                                                              }
{ Type conversion                                                              }
{                                                                              }
Function StrToIntDefault (const S : String; const Default : Integer) : Integer;
  Begin
    try
      Result := StrToInt (S);
    except
      Result := Default;
    end;
  End;

Function StrToFloatDefault (const S : String; const Default : Extended) : Extended;
  Begin
    try
      Result := StrToFloat (S);
    except
      Result := Default;
    end;
  End;

Function BooleanToStr (const B : Boolean) : String;
  Begin
    Result := Cond (B, 'True', 'False');
  End;

Function StrToBoolean (const S : String) : Boolean;
  Begin
    Result := LowerCase (S) = 'true';
  End;

Function ExceptionToStr (const E : Exception) : String;
  Begin
    Result := E.ClassName + ' (' + E.Message + ')';
  End;



{                                                                              }
{ Path functions                                                               }
{                                                                              }
Function PathWithSlashAtEnd (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    if (Path <> '') and (Path [Length (Path)] <> SlashCh) then
      Result := Path + SlashCh else
      Result := Path;
  End;

Function PathWithoutSlashAtEnd (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    if (Path <> '') and (Path [Length (Path)] = SlashCh) then
      Result := Copy (Path, 1, Length (Path) - 1) else
      Result := Path;
  End;

Function UnixPathToWindowsPath (const Path : String) : String;
  Begin
    Result := Replace ('/', '\', Replace (['\', ':', '"', '<', '>', '|'], '_', Path));
  End;

{ Network path:  \\      --->   /                                              }
{ Device path:   \\.\    --->   /                                              }
{ Drive letter:  X:      --->   X/                                             }
Function WindowsPathToUnixPath (const Path : String) : String;
  Begin
    Result := Path;
    if Length (Result) >= 2 then
      if Result [2] = ':' then // drive letter
        Result [2] := '/' else
      if (Result [1] = '\') and (Result [2] = '\') then
        if (Length (Result) >= 3) and (Result [3] = '.') then
          Delete (Result, 1, 3) else    // device path
          Delete (Result, 1, 1);        // network path
    Result := Replace ('\', '/', Replace (['/', ':', '"', '<', '>', '|'], '_', Path));
  End;

Function PathLeftElement (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    Result := CopyBefore (Path, SlashCh, True);
  End;

Function PathRightElement (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    Result := CopyFrom (Path, PosPrev (SlashCh, Path) + 1);
  End;

Function PathWithoutLeftElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
  Begin
    Result := CopyAfter (Path, SlashCh, False);
  End;

Function PathWithoutRightElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
  Begin
    Result := CopyBefore (Path, SlashCh, False);
  End;

{                                                                              }
{ Locked integer manipulation                                                  }
{ Code by Azret Botash.                                                        }
{   LockedAdd, LockedDec, LockedSub, LockedInc does operation with Value on    }
{     Target and returns the result.                                           }
{   LockedCompareExchange compares Target with Comp and if equal sets Target   }
{     to Exch.                                                                 }
{   LockedExchange set current value of Target to Value and returns old value  }
{     of Target.                                                               }
{                                                                              }
Function LockedAdd (var Target : Integer; const Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, EDX
      LOCK XADD [ECX], EAX
      ADD     EAX, EDX
  End;

Function LockedCompareExchange (var Target : Integer; const Exch, Comp : Integer) : Integer; assembler; register;
  Asm
      XCHG    EAX, ECX
      LOCK CMPXCHG [ECX], EDX
  End;

Function LockedDec (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, -1
      LOCK XADD [ECX], EAX
      DEC     EAX
  End;

Function LockedExchange (var Target : Integer; Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, EDX
      LOCK XCHG [ECX], EAX
  End;

Function LockedExchangeAdd (var Target : Integer; Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, EDX
      LOCK XADD [ECX], EAX
  End;

Function LockedExchangeDec (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, -1
      LOCK XADD [ECX], EAX
  End;

Function LockedExchangeInc (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, 1
      LOCK XADD [ECX], EAX
  End;

Function LockedExchangeSub (var Target : Integer; Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      NEG     EDX
      MOV     EAX, EDX
      LOCK XADD [ECX], EAX
  End;

Function LockedInc (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, 1
      LOCK XADD [ECX], EAX
      INC     EAX
  End;

Function LockedSub (var Target : Integer; const Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      NEG     EDX
      MOV     EAX, EDX
      LOCK XADD [ECX], EAX
      ADD     EAX, EDX
  End;

end.

